/*
 * Copyright:      Fabian Maximilian Thiele  mailto:me@apetech.de
 * Author:         Fabian Maximilian Thiele
 * Remarks:        this Copyright must be included
 * known Problems: none
 * Version:        1.0
 * Description:    Scheduler mit dynamischer Aufgaben-Liste
 * 
 */

#include <avr/io.h>
#include <avr/signal.h>
#include <avr/interrupt.h>
#include <avr/wdt.h>
#include <inttypes.h>
#include <stdlib.h>

#include "scheduler.h"
#include "rtc.h"

// Scheduler-Zeit
volatile date_t schedulerTime;

// Anzahl der Eintrge in der Scheduler-Liste
volatile uint8_t schedulerLength = 0;

// Pointer auf erstes Element der Scheduler Queue
schedulerTask_t* schedulerQueuePtr = NULL;


/**
 * Initialisiert den Scheduler.
 */
void schedulerInit(void) {
	schedulerTime.subSec = 0;
	schedulerTime.sec = 0;
	schedulerTime.min = 0;
	schedulerTime.hour = 0;
	schedulerTime.day = 0;
	schedulerTime.month = 0;
	schedulerTime.year = 2005;
}

/*
 * vergleicht eine Zeit mit der Scheduler Zeit. Wird gebraucht um zu gucken
 * ob ein Task ausgefhrt werden soll oder nicht.
 */
int8_t schedulerCompToSystime(date_t* date) {
	if(schedulerTime.year > (*date).year) return -1;
	else if(schedulerTime.year < (*date).year) return 1;
	
	else if(schedulerTime.month > (*date).month) return -1;
	else if(schedulerTime.month < (*date).month) return 1;
	
	else if(schedulerTime.day > (*date).day) return -1;
	else if(schedulerTime.day < (*date).day) return 1;
	
	else if(schedulerTime.hour > (*date).hour) return -1;
	else if(schedulerTime.hour < (*date).hour) return 1;
	
	else if(schedulerTime.min > (*date).min) return -1;
	else if(schedulerTime.min < (*date).min) return 1;
	
	else if(schedulerTime.sec > (*date).sec) return -1;
	else if(schedulerTime.sec < (*date).sec) return 1;
	
	else if(schedulerTime.subSec > (*date).subSec) return -1;
	else if(schedulerTime.subSec < (*date).subSec) return 1;
	
	return 0;
}

/**
 * Fgt der Scheduler-Liste einen Eintrag hinzu.
 * Dafr werden der Ausfhrungszeitpunkt, ein benutzerdefinierbarer 16Bit int
 * und die Callback-Funktion bergeben.
 * Der Integer wird der Callback-Funktion als Parameter bergeben und kann
 * einen beliebeigen Wert haben.
 */
uint16_t schedulerAddTask(schedulerCallback_t task, date_t* taskDate, uint16_t taskParam) {
	static uint16_t taskId = 0;
	uint8_t again = 0;
	
	schedulerTask_t* ptr = malloc(sizeof(schedulerTask_t));
	// 0 returnen, falls kein Platz mehr im Speicher
	if(ptr == NULL) {
		return 0;
	}
	
	(*ptr).nextTaskPtr = NULL;
	(*ptr).task = task;
	(*ptr).taskDate = *taskDate;
	(*ptr).taskParam = taskParam;
	
	// taskId fr nchsten Eintrag inkrementieren und bei Overflow auf 1 setzen.
	if(!(++taskId)) taskId++;
	
	schedulerTask_t* ptr1 = NULL;
	schedulerTask_t* lptr = NULL;
	// Schon Eintrge vorhanden?
	if(schedulerLength) {
		// Prfen ob die Id schon vorhanden ist. Nicht gerade effizient, aber
		// es drfte nicht oft vorkommen das eine ID bereits vergeben ist.
		do {
			ptr1 = schedulerQueuePtr;
			while(ptr1 != NULL) {
				if((*ptr1).taskId == taskId) {
					// ID schon vorhanden. Inkrementieren und nochmal testen
					taskId++;
					again = 1;
				}
				lptr = ptr1;
				ptr1 = (*ptr1).nextTaskPtr;
			}
		} while(again);
	
		// Eintrag ans Ende der Liste anfgen
		(*ptr).taskId = taskId;
		(*lptr).nextTaskPtr = ptr;
		
	} else {
		// Wenn noch keine Eintrge vorhanden...
		(*ptr).taskId = taskId;
		schedulerQueuePtr = ptr;
	}
	
	schedulerLength++;
	
	return taskId;
}

/**
 * Prft ob Eintrge in der Liste ausgefhrt werden sollen und fhrt diese
 * dann auch aus.
 * Diese Funktion sollte regelmig aus der Main-Loop heraus aufgerufen werden.
 * Falls sehr zeitkritische Aufgaben bearbeitet werden sollen, kann die
 * Funktion auch aus der RTC-Timer ISR heraus aufgerufen werden, allerdings
 * sollte dabei beachtet werden, dass dann auch die Callback-Funktion aus
 * dem Interrupt heraus ausgefhrt wird. Auf jeden Fall sollten in diesem Falle
 * die Interrupts wieder freigegeben werden.
 */
void schedulerCheck(void) {
	// Erstes Element der Liste laden
	schedulerTask_t* ptr = schedulerQueuePtr;
	schedulerTask_t* lptr = NULL;
	
	// Liste bis zum Ende durchtickern
	for(uint8_t i=0; i<schedulerLength; i++) {
		
		if(schedulerCompToSystime(&(*ptr).taskDate) < 0) {
			if((*ptr).task((*ptr).taskParam, &(*ptr).taskDate) == SCHEDULER_TERMINATE_TASK) {
				schedulerTask_t* ptr1 = (*ptr).nextTaskPtr;
				
				if(lptr != NULL) {
					(*lptr).nextTaskPtr = (*ptr).nextTaskPtr;
				} else {
					schedulerQueuePtr = NULL;
				}
				
				free(ptr);
				ptr = ptr1;
				schedulerLength--;
			}
		}
		
		lptr = ptr;
		ptr = (*ptr).nextTaskPtr;
	}
}

/**
 * Sucht einen Scheduler-Eintrag an Hand seiner ID und setzt seinen
 * Ausfhrungszeitpunkt auf den bergebenen Wert.
 */
void schedulerEditTaskTime(uint16_t taskId, date_t* taskDate) {
	schedulerTask_t* ptr = schedulerQueuePtr;
	
	if(!taskId) return;
	
	// Liste durchtickern und ID suchen
	cli();
	while(ptr != NULL) {
		if((*ptr).taskId == taskId) {
			// Task mit passender ID editieren und wieder raus
			(*ptr).taskDate = *taskDate;
			sei();
			return;
		}
		ptr = (*ptr).nextTaskPtr;
	}
	sei();
}

/**
 * Sucht einen Scheduler-Eintrag an Hand seiner ID und setzt seinen
 * Callback-Parameter auf den bergebenen Wert.
 */
void schedulerEditTaskParam(uint16_t taskId, uint16_t taskParam) {
	schedulerTask_t* ptr = schedulerQueuePtr;
	
	if(!taskId) return;
	
	// Liste durchtickern und ID suchen
	cli();
	while(ptr != NULL) {
		if((*ptr).taskId == taskId) {
			// Task mit passender ID editieren und wieder raus
			(*ptr).taskParam = taskParam;
			sei();
			return;
		}
		ptr = (*ptr).nextTaskPtr;
	}
	sei();
}

/**
 * Sucht einen Scheduler-Eintrag an Hand seiner ID und fhrt ihn aus.
 * Der Ausfhrungszeitpunkt wird hierbei nicht beachtet.
 * Wenn die Callback-Funktion SCHEDULER_TERMINATE_TASK zurckgibt, wird
 * der Eintrag auch aus der Liste gelscht.
 */
void schedulerForceTask(uint16_t taskId) {
	if(!taskId) return;
	
	// Liste durchtickern und ID suchen
	cli();
	
	// Erstes Element der Liste laden
	schedulerTask_t* ptr = schedulerQueuePtr;
	schedulerTask_t* lptr = NULL;
	
	if(!taskId) return;
	
	// Liste durchtickern und ID suchen
	cli();
	while(ptr != NULL) {
		if((*ptr).taskId == taskId) {
			sei();
			if(!(*ptr).task((*ptr).taskParam, &(*ptr).taskDate)) {	
				if(lptr != NULL) {
					(*lptr).nextTaskPtr = (*ptr).nextTaskPtr;
				} else {
					schedulerQueuePtr = NULL;
				}
				
				free(ptr);
				schedulerLength--;
			}
			return;
		}
		lptr = ptr;
		ptr = (*ptr).nextTaskPtr;
	}
	sei();
}

/**
 * Sucht einen Scheduler-Eintrag an Hand seiner ID und lscht ihn.
 */
void schedulerRemoveTask(uint16_t taskId) {
	schedulerTask_t* ptr = schedulerQueuePtr;
	schedulerTask_t* lptr = NULL;
	
	if(!taskId) return;
	
	// Liste durchtickern und ID suchen
	cli();
	while(ptr != NULL) {
		if((*ptr).taskId == taskId) {
			// Task mit passender ID entfernen und wieder raus
			if(lptr != NULL) {
				(*lptr).nextTaskPtr = (*ptr).nextTaskPtr;
			} else {
				schedulerQueuePtr = NULL;
			}
				
			free(ptr);
			schedulerLength--;
			sei();
			return;
		}
		lptr = ptr;
		ptr = (*ptr).nextTaskPtr;
	}
	sei();
}
